/*generic libraries*/
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <string.h>
#include <signal.h>

/*broccoli !*/
#include <broccoli.h>

/*prelude*/
#include <libprelude/prelude.h>

/*our headers*/
#include "prelude.h"
#include "utils.h"

/*default socket settings*/
char *port_str = "47759";
/*default prelude profile*/
char *profile = "bro2prelude";

/*are we running or whould we stop*/
static int running = 0;

static void
usage(void)
{
  printf("broprelude - sends alerts via prelude library.\n"
	 "USAGE: broprelude [-h|-?] [-p port]\n"
	 "   -h|-?       This message.\n"
	 "   -a <prelude analyzer>      prelude profile to use.\n"
	 "   -p <port>   Port where to be contacted.\n");

  exit(0);
}

/*when we catch a signal, we stop, smoothly if we are running the main loop*/
static void sigHandler(int sigNum)
{
  if (running)
  {
    running = 0;
  }
  else
  {
    exit(-1);
  }
}

/*get the file descriptor of our socket awaiting connections*/
int start_listen(int port)
{
  int fd = 0;
  struct sockaddr_in server;
  const int turn_on = 1;

  fd = socket(PF_INET, SOCK_STREAM, 0);
  if ( fd < 0 )  
	{
	printf("can't create listen socket: %s\n", strerror(errno));
	exit(-1);
	}

  // Set SO_REUSEADDR.
  if ( setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &turn_on, sizeof(turn_on)) < 0 )
	{
	printf("can't set SO_REUSEADDR: %s\n", strerror(errno));
	exit(-1);
	}

  bzero(&server, sizeof(server));
  server.sin_family = AF_INET;
  server.sin_port = htons(port);
  server.sin_addr.s_addr = 0;

  if ( bind(fd, (struct sockaddr*) &server, sizeof(server)) < 0 )
	{
	printf("can't bind to port %d: %s\n", port, strerror(errno));
	exit(-1);
	}

  if ( listen(fd, 50) < 0 )
	{
	printf("can't listen: %s\n", strerror(errno));
	exit(-1);
	}

  return fd;
}

/*make the set of file descriptors for select*/
static int makeFDSet(fd_set * fds, BP_LinkedListFDBC_t * fdList)
{
  BP_LinkedListFDBC_t * current = fdList;
  int fdMax = -1;

  FD_ZERO(fds);
  while (current)
  {
    if (current->fd > fdMax)
    {
      fdMax = current->fd;
    }
    FD_SET(current->fd, fds);
    current = current->next;
  }
  return fdMax;
}

static void handleReceivedData(fd_set * fds, BP_LinkedListFDBC_t * fdList)
{
  BP_LinkedListFDBC_t * current = fdList;
  int fd = -1;
  BroConn *bc = NULL;
  struct sockaddr_in client;
  socklen_t len = sizeof(client);

  while (current)
  {
    if (FD_ISSET(current->fd, fds))
    {
      if (current->bc)
      {
        /*a Bro sent something*/
        bro_conn_process_input(current->bc);
        if (!bro_conn_alive(current->bc))
        {
          printf("Bro disconnected %d\n", current->fd);
          removeFDBClist(fdList, current);
        }
      }
      else
      {
        /*a new Bro is connecting*/
        fd = accept(current->fd, (struct sockaddr*) &client, &len);
        if ( fd < 0 )
        {
          printf("can't accept: %s\n", strerror(errno));
          exit(-1);
        }
        bc = bro_conn_new_socket(fd, BRO_CFLAG_NONE);
        if ( ! bc )
        {
          printf("can't create connection form fd\n");
          exit(-1);
        }
        bro_event_registry_add(bc, "preludeAlert", (BroEventFunc) bro_preludeAlert, NULL);
        bro_conn_set_class(bc, "prelude");
        if (! bro_conn_connect(bc))
        {
          printf("Could not connect to Bro.\n");
          exit(-1);
        }
        printf("Bro connected %d\n", fd);
        addFDBClist(fdList, fd, bc);
      }
    }
    current = current->next;
  }
}

int
main(int argc, char **argv)
{
  int opt, port;
  extern char *optarg;
  extern int optind;
  int fd = -1;
  int fdConn = -1;
  fd_set fds;
  sigset_t setSigTerm;
  BP_LinkedListFDBC_t * fdList = NULL;
  struct timeval timeout;

/*options parsing*/
  while ( (opt = getopt(argc, argv, "a:p:h?")) != -1)
    {
      switch (opt)
	{
	case 'h':
	case '?':
	  usage();
	  
	case 'p':
	  port_str = optarg;
	  break;

	case 'a':
	  profile = optarg;
	  break;

	default:
	  usage();
	}
    }
  argc -= optind;
  argv += optind;
  port = strtol(port_str, NULL, 0);
  if (errno == ERANGE)
    {
      printf("Please provide a port number with -p.\n");
      exit(-1);
    }

/*signal handling : catching SIGTERM to stop smoothly*/
  if (sigemptyset(&setSigTerm))
  {
    fprintf(stderr, "setting signal mask error\n");
    exit(-1);
  }
  if (sigaddset(&setSigTerm, SIGTERM))
  {
    fprintf(stderr, "adding signal mask error\n");
    exit(-1);
  }
  if (signal(SIGTERM, &sigHandler) == SIG_ERR)
  {
    fprintf(stderr, "setting signal handler error\n");
    exit(-1);
  }

/* Broccoli initialisation */
  bro_init(NULL);
  fdConn = start_listen(port);
  fdList = newFDBClist(fdConn);

/* Prelude initialisation */
  BP_initPrelude(profile);

  running = 1;
  /* Enter processing loop */
  while ( running )
    {
      fd = makeFDSet(&fds, fdList);
      /*1 second timeout : the ending of the process must not last too long*/
      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
      /*block signals while selecting*/
      if (sigprocmask(SIG_BLOCK, &setSigTerm, NULL))
      {
        fprintf(stderr, "blocking signals error\n");
        exit(-1);
      }
      if ( select(fd + 1, &fds, NULL, NULL, &timeout) < 0 )
        {
          printf("can't select: %s\n", strerror(errno));
          break;
        }
      handleReceivedData(&fds, fdList);
      if (sigprocmask(SIG_UNBLOCK, &setSigTerm, NULL))
      {
        fprintf(stderr, "unblocking signals error\n");
        exit(-1);
      }
      /*FIXME timer waken up too often*/
      prelude_timer_wake_up();
      sleep(1);
      prelude_timer_wake_up();
    }

  printf("ending bro2prelude\n");
  /* cleaning */
  cleanFDBClist(fdList);
  BP_cleanPrelude();

  return 0;
}
