#include <broccoli.h>
#include <libprelude/prelude.h>
#include <math.h>

/*the global variable of our prelude client*/
static prelude_client_t * client;

/*initialisation*/
void BP_initPrelude(const char * profile)
{
	int ret = 0;

	ret = prelude_init(NULL, NULL);
	if (ret < 0)
	{
		prelude_perror(ret, "error init libprelude");
		exit(-1);
	}

	ret = prelude_client_new(&client, profile);
	if (!client || ret < 0)
	{
		prelude_perror(ret, "error creation client prelude");
		exit(-1);
	}

	ret = prelude_client_set_flags(client, PRELUDE_CLIENT_FLAGS_CONNECT | PRELUDE_CLIENT_FLAGS_HEARTBEAT | PRELUDE_CLIENT_FLAGS_AUTOCONFIG);
	if ( ret < 0 ) 
	{
		fprintf(stderr, "Unable to set asynchronous send and timer.\n");
		exit(-1);
	}

	ret = prelude_client_start(client);
	if (ret < 0)
	{
		prelude_perror(ret, "error starting client prelude");
		prelude_client_destroy(client, PRELUDE_CLIENT_EXIT_STATUS_FAILURE);
		exit(-1);
	}
}

/*basic alert creation*/
static int createAlert(idmef_message_t ** idmef, idmef_alert_t ** alert)
{
	int ret = 0;
	idmef_time_t *time;

	ret = idmef_message_new(idmef);
	if ( ret < 0 ) 
	{
		prelude_perror(ret, "Unable to create the IDMEF message");
		return -1;
	}
	
	ret = idmef_message_new_alert(*idmef, alert);
	if ( ret < 0 ) 
	{
		prelude_perror(ret, "unable to create IDMEF alert");
		idmef_message_destroy(*idmef);
		return -1;
	}

	/*set create time*/
	ret = idmef_time_new_from_gettimeofday(&time);
	if ( ret < 0 )
	{
		prelude_perror(ret, "unable to get current time");
		idmef_message_destroy(*idmef);
		return -1;
	}
        idmef_alert_set_create_time(*alert, time);

	return 0;
}

/*macro to check errors*/
#define BP_IDMEF_IF_ASSERTED(x) if (x<0)\
{\
	fprintf(stderr, "prelude failed at %s[%d]\n%s\n", __FILE__, __LINE__, #x);\
}\
else

/*fills the classification field*/
static void createClassification(idmef_alert_t * alert, const char * nom)
{
	prelude_string_t * ident;
	prelude_string_t * text;
	idmef_classification_t * class;

	BP_IDMEF_IF_ASSERTED (idmef_alert_new_classification(alert, &class))
	{
		BP_IDMEF_IF_ASSERTED (idmef_classification_new_text(class, &text))
		{
			prelude_string_set_dup(text, nom);
		}
		BP_IDMEF_IF_ASSERTED (idmef_classification_new_ident(class, &ident))
		{
			prelude_string_set_dup(ident, nom);
		}
	}
}

#define IANA_PROTONUMER_TCP 6
#define IANA_PROTONUMER_UDP 17
#define IANA_PROTONUMER_ICMP 1

/*fills the source and target fields*/
static void createSourceAndTarget(idmef_alert_t * alert, BroRecord *conn_id)
{
	idmef_node_t * node;
	idmef_service_t * service;
	idmef_source_t * source;
	idmef_target_t * target;

	BP_IDMEF_IF_ASSERTED (idmef_alert_new_source(alert, &source, IDMEF_LIST_APPEND))
	{
		BP_IDMEF_IF_ASSERTED (idmef_source_new_node(source, &node))
		{
			idmef_address_t * address;
			BP_IDMEF_IF_ASSERTED (idmef_node_new_address(node, &address, IDMEF_LIST_APPEND))
			{
				prelude_string_t * preludeAddr;
				uint32_t * ip = bro_record_get_named_val(conn_id, "orig_h", NULL);

				BP_IDMEF_IF_ASSERTED (idmef_address_new_address(address, &preludeAddr))
				{
					prelude_string_sprintf(preludeAddr, "%d.%d.%d.%d", (ntohl(*ip) & 0xFF000000) >> 24, 
						(ntohl(*ip) & 0x00FF0000) >> 16, (ntohl(*ip) & 0x0000FF00) >> 8, (ntohl(*ip) & 0x000000FF));
				}
				idmef_address_set_category(address, IDMEF_ADDRESS_CATEGORY_IPV4_ADDR);
;
			}
		}
		BP_IDMEF_IF_ASSERTED (idmef_source_new_service(source, &service))
		{
			prelude_string_t * preludeProto;
			BroPort * port = bro_record_get_named_val(conn_id, "orig_p", NULL);

			idmef_service_set_port(service, port->port_num);
			BP_IDMEF_IF_ASSERTED (idmef_service_new_protocol(service, &preludeProto))
			{
				if (port->port_proto == IANA_PROTONUMER_TCP)
				{
					prelude_string_sprintf(preludeProto, "tcp");
				}
				else if (port->port_proto == IANA_PROTONUMER_UDP)
				{
					prelude_string_sprintf(preludeProto, "udp");
				}
				else if (port->port_proto == IANA_PROTONUMER_ICMP)
				{
					prelude_string_sprintf(preludeProto, "icmp");
				}
				else
				{
					prelude_string_sprintf(preludeProto, "unknown%d", port->port_proto);
				}
			}
		}
	}

	/*FIXME code above should be written only once in a function*/
	BP_IDMEF_IF_ASSERTED (idmef_alert_new_target(alert, &target, IDMEF_LIST_APPEND))
	{
		BP_IDMEF_IF_ASSERTED (idmef_target_new_node(target, &node))
		{
			idmef_address_t * address;
			BP_IDMEF_IF_ASSERTED (idmef_node_new_address(node, &address, IDMEF_LIST_APPEND))
			{
				prelude_string_t * preludeAddr;
				uint32_t * ip = bro_record_get_named_val(conn_id, "resp_h", NULL);

				BP_IDMEF_IF_ASSERTED (idmef_address_new_address(address, &preludeAddr))
				{
					prelude_string_sprintf(preludeAddr, "%d.%d.%d.%d", (ntohl(*ip) & 0xFF000000) >> 24,
						(ntohl(*ip) & 0x00FF0000) >> 16, (ntohl(*ip) & 0x0000FF00) >> 8, (ntohl(*ip) & 0x000000FF));
				}
				idmef_address_set_category(address, IDMEF_ADDRESS_CATEGORY_IPV4_ADDR);
;
			}
		}
		BP_IDMEF_IF_ASSERTED (idmef_target_new_service(target, &service))
		{
			prelude_string_t * preludeProto;
			BroPort * port = bro_record_get_named_val(conn_id, "resp_p", NULL);

			idmef_service_set_port(service, port->port_num);
			BP_IDMEF_IF_ASSERTED (idmef_service_new_protocol(service, &preludeProto))
			{
				if (port->port_proto == IANA_PROTONUMER_TCP)
				{
					prelude_string_sprintf(preludeProto, "tcp");
				}
				else if (port->port_proto == IANA_PROTONUMER_UDP)
				{
					prelude_string_sprintf(preludeProto, "udp");
				}
				else if (port->port_proto == IANA_PROTONUMER_ICMP)
				{
					prelude_string_sprintf(preludeProto, "icmp");
				}
				else
				{
					prelude_string_sprintf(preludeProto, "unknown%d", port->port_proto);
				}
			}
		}
	}
}

/*callback on a table containing the different additional datas*/
static int createAdditionalData(void *key, void *val, void * user_data)
{
	BroString *keystring = (BroString *) key;
/*maybe we should have something more generic here*/
	BroString *valstring = (BroString *) val;
	idmef_alert_t * alert = (idmef_alert_t *) user_data;
	prelude_string_t * addMeaning;
	idmef_additional_data_t * additionalData;

	if (valstring->str_len == 0)
	{
		return 1;
	}

	BP_IDMEF_IF_ASSERTED (idmef_alert_new_additional_data(alert, &additionalData, IDMEF_LIST_APPEND))
	{
		BP_IDMEF_IF_ASSERTED (idmef_additional_data_new_meaning(additionalData, &addMeaning))
		{
			prelude_string_sprintf(addMeaning, "%s", keystring->str_val);
		}
		idmef_additional_data_set_type(additionalData, IDMEF_ADDITIONAL_DATA_TYPE_STRING);
		idmef_additional_data_set_string_ref(additionalData, (char *) valstring->str_val);
	}
	return 1;
}

/*set a value in an idmef path*/
static void set_idmef_path(idmef_message_t *message, const char *pathname, const char *value)
{
        int ret;
        idmef_value_t *val;
        idmef_path_t *path;
        ret = idmef_path_new(&path, pathname);
        if ( ret < 0 )
                return ;
        ret = idmef_value_new_from_path(&val, path, value);
        if ( ret < 0 ) {
                idmef_path_destroy(path);
                return ;
        }
        ret = idmef_path_set(path, message, val);
        idmef_value_destroy(val);
        idmef_path_destroy(path);
        return ;
}

/*set a generic idmef field*/
static int createIDMEFfield(void *key, void *val, void * user_data)
{
	BroString *keystring = (BroString *) key;
	BroString *valstring = (BroString *) val;
	idmef_message_t * idmef = (idmef_message_t *) user_data;

	set_idmef_path(idmef, (char *) keystring->str_val, (char *) valstring->str_val);
	return 1;
}

/*send the alert via prelude*/
static void BP_sendAlert(idmef_message_t * idmef, idmef_alert_t * alert)
{
	int ret = 0;
	idmef_time_t *time;

	/*set analyzertime*/
	ret = idmef_time_new_from_gettimeofday(&time);
	if ( ret < 0 )
	{
		prelude_perror(ret, "unable to get current time");
	}
        idmef_alert_set_analyzer_time(alert, time);
	prelude_client_send_idmef(client, idmef);
	idmef_message_destroy(idmef);
}

/*creation of a prelude alert*/
void bro_preludeAlert(BroConn *conn, void *data, BroRecord * broAlert)
{
	double * detectTime;
	BroString *classification;
	BroRecord *conn_id;
	BroTable *fields;
	BroTable *addData;

	struct timeval tv;
	double fracpart, intpart;

	idmef_alert_t * alert;
	idmef_message_t * idmef;
	idmef_time_t * time;

	if (createAlert(&idmef, &alert))
	{
		fprintf(stderr, "alert creation error\n");
		return;
	}

	detectTime = bro_record_get_named_val(broAlert, "ts", NULL);
	fracpart = modf(*detectTime, &intpart);
	tv.tv_sec = (int) intpart;
	tv.tv_usec = (int) (fracpart * 1000000);
	BP_IDMEF_IF_ASSERTED (idmef_time_new_from_timeval(&time, &tv))
	{
		idmef_alert_set_detect_time(alert, time);
	}

	classification = bro_record_get_named_val(broAlert, "classification", NULL);
	createClassification(alert, (char *) classification->str_val);

	conn_id = bro_record_get_named_val(broAlert, "id", NULL);
	if (conn_id)
	{
		createSourceAndTarget(alert, conn_id);
	}

	fields = bro_record_get_named_val(broAlert, "fields", NULL);
	if (fields)
	{
		bro_table_foreach(fields, createIDMEFfield, idmef);
	}

	addData = bro_record_get_named_val(broAlert, "additional_data", NULL);
	if (addData)
	{
		bro_table_foreach(addData, createAdditionalData, alert);
	}

	BP_sendAlert(idmef, alert);
}

/*cleaning*/
void BP_cleanPrelude()
{
	prelude_client_destroy(client, PRELUDE_CLIENT_EXIT_STATUS_SUCCESS);
	prelude_deinit();
}

